# mypy: ignore-errors
from typing import Any

import pandas as pd

from vivarium import Component
from vivarium.framework.engine import Builder


class TreatmentIntervention(Component):
    CONFIGURATION_DEFAULTS: dict[str, Any] = {
        "intervention": {
            "effect_size": 0.5,
        }
    }

    ##############
    # Properties #
    ##############

    @property
    def configuration_defaults(self) -> dict[str, Any]:
        return {self.intervention: self.CONFIGURATION_DEFAULTS["intervention"]}

    #####################
    # Lifecycle methods #
    #####################

    def __init__(self, intervention: str, affected_value: str):
        super().__init__()
        self.intervention = intervention
        self.affected_value = affected_value

    # noinspection PyAttributeOutsideInit
    def setup(self, builder: Builder) -> None:
        effect_size = builder.configuration[self.intervention].effect_size
        self.effect_size = builder.value.register_value_producer(
            f"{self.intervention}.effect_size",
            source=lambda index: pd.Series(effect_size, index=index),
        )
        builder.value.register_value_modifier(
            self.affected_value,
            modifier=self.intervention_effect,
            required_resources=[self.effect_size],
        )

    ##################################
    # Pipeline sources and modifiers #
    ##################################

    def intervention_effect(self, index: pd.Index, value: pd.Series) -> pd.Series:
        return value * (1 - self.effect_size(index))



from __future__ import annotations

from collections.abc import Callable, Iterator, Sequence
from typing import TYPE_CHECKING

from vivarium.framework.resource.exceptions import ResourceError
from vivarium.framework.resource.resource import Resource

if TYPE_CHECKING:
    from vivarium.framework.population import SimulantData


class ResourceGroup:
    """Resource groups are the nodes in the resource dependency graph.

    A resource group represents the pool of resources produced by a single
    callable and all the dependencies necessary to produce that resource.
    When thinking of the dependency graph, this represents a vertex and
    all in-edges.  This is a local-information representation that can be
    used to construct the entire dependency graph once all resources are
    specified.

    """

    def __init__(
        self, initialized_resources: Sequence[Resource], dependencies: Sequence[Resource]
    ):
        """Create a new resource group.

        Parameters
        ----------
        initialized_resources
            The resources initialized by this resource group's initializer.
        dependencies
            The resources this resource group's initializer depends on.

        Raises
        ------
        ResourceError
            If the resource group is not well-formed.
        """
        if not initialized_resources:
            raise ResourceError("Resource groups must have at least one resource.")

        if len(set(r.component for r in initialized_resources)) != 1:
            raise ResourceError("All initialized resources must have the same component.")

        if len(set(r.resource_type for r in initialized_resources)) != 1:
            raise ResourceError("All initialized resources must be of the same type.")

        self.component = initialized_resources[0].component
        """The component or manager that produces the resources in this group."""
        self.type = initialized_resources[0].resource_type
        """The type of resource in this group."""
        self.is_initialized = initialized_resources[0].is_initialized
        """Whether this resource group contains initialized resources."""
        self._dependencies = dependencies
        self.resources = {r.resource_id: r for r in initialized_resources}
        """A dictionary of resources produced by this group, keyed by resource_id."""

    @property
    def names(self) -> list[str]:
        """The long names (including type) of all resources in this group."""
        return list(self.resources)

    @property
    def initializer(self) -> Callable[[SimulantData], None]:
        """The method that initializes this group of resources."""
        # TODO [MIC-5452]: all resource groups should have a component
        if not self.component:
            raise ResourceError(f"Resource group {self} does not have an initializer.")
        return self.component.on_initialize_simulants

    @property
    def dependencies(self) -> list[str]:
        """The long names (including type) of dependencies for this group."""
        return [dependency.resource_id for dependency in self._dependencies]

    def __iter__(self) -> Iterator[str]:
        return iter(self.names)

    def __repr__(self) -> str:
        resources = ", ".join(self)
        return f"ResourceProducer({resources})"

    def __str__(self) -> str:
        resources = ", ".join(self)
        return f"({resources})"

    def get_resource(self, resource_id: str) -> Resource:
        """Get a resource by its resource_id."""
        return self.resources[resource_id]



import itertools
import math
import re
from collections.abc import Callable
from datetime import timedelta
from typing import Any
from unittest.mock import Mock

import numpy as np
import pandas as pd
import pytest
from layered_config_tree.main import LayeredConfigTree
from loguru import logger
from pandas.core.groupby.generic import DataFrameGroupBy
from pytest_mock import MockerFixture

from tests.framework.results.helpers import (
    BASE_POPULATION,
    FAMILIARS,
    HOUSE_CATEGORIES,
    NAME,
    NAME_COLUMNS,
    sorting_hat_serial,
    sorting_hat_vectorized,
    verify_stratification_added,
)
from vivarium.framework.event import Event
from vivarium.framework.lifecycle import lifecycle_states
from vivarium.framework.results import VALUE_COLUMN
from vivarium.framework.results.context import ResultsContext
from vivarium.framework.results.observation import AddingObservation, ConcatenatingObservation
from vivarium.framework.results.stratification import Stratification, get_mapped_col_name
from vivarium.framework.values import Pipeline
from vivarium.types import ScalarMapper, VectorMapper


def _aggregate_state_person_time(x: pd.DataFrame) -> float:
    """Helper aggregator function for observation testing"""
    return len(x) * (28 / 365.25)


@pytest.fixture
def event() -> Event:
    return Event(
        name=lifecycle_states.COLLECT_METRICS,
        index=pd.Index([0]),
        user_data={},
        time=0,
        step_size=1,
    )


@pytest.mark.parametrize(
    "mapper, is_vectorized",
    [
        (sorting_hat_vectorized, True),
        (sorting_hat_serial, False),
    ],
    ids=["vectorized_mapper", "non-vectorized_mapper"],
)
def test_add_stratification_mappers(
    mapper: VectorMapper | ScalarMapper, is_vectorized: bool, mocker: MockerFixture
) -> None:
    ctx = ResultsContext()
    mocker.patch.object(ctx, "excluded_categories", {})
    pipeline = Pipeline("grade")

    assert NAME not in ctx.stratifications

    ctx.add_stratification(
        name=NAME,
        requires_columns=NAME_COLUMNS,
        requires_values=[pipeline],
        categories=HOUSE_CATEGORIES,
        excluded_categories=None,
        mapper=mapper,
        is_vectorized=is_vectorized,
    )
    assert verify_stratification_added(
        stratifications=ctx.stratifications,
        name=NAME,
        requires_columns=NAME_COLUMNS,
        requires_values=[pipeline],
        categories=HOUSE_CATEGORIES,
        excluded_categories=[],
        mapper=mapper,
        is_vectorized=is_vectorized,
    )


@pytest.mark.parametrize(
    "excluded_categories",
    [
        [],
        HOUSE_CATEGORIES[:1],
        HOUSE_CATEGORIES[:2],
        HOUSE_CATEGORIES[:3],
    ],
    ids=[
        "no_excluded_categories",
        "one_excluded_category",
        "two_excluded_categories",
        "all_but_one_excluded_categories",
    ],
)
def test_add_stratification_excluded_categories(
    excluded_categories: list[str], mocker: MockerFixture
) -> None:
    ctx = ResultsContext()
    builder = mocker.Mock()
    builder.configuration.stratification = LayeredConfigTree(
        {"default": [], "excluded_categories": {NAME: excluded_categories}}
    )
    builder.logging.get_logger.return_value = logger
    ctx.setup(builder)

    assert NAME not in ctx.stratifications

    ctx.add_stratification(
        name=NAME,
        requires_columns=NAME_COLUMNS,
        requires_values=[],
        categories=HOUSE_CATEGORIES,
        excluded_categories=excluded_categories,
        mapper=sorting_hat_vectorized,
        is_vectorized=True,
    )

    assert verify_stratification_added(
        stratifications=ctx.stratifications,
        name=NAME,
        requires_columns=NAME_COLUMNS,
        requires_values=[],
        categories=HOUSE_CATEGORIES,
        excluded_categories=excluded_categories,
        mapper=sorting_hat_vectorized,
        is_vectorized=True,
    )


@pytest.mark.parametrize(
    "name, categories, excluded_categories, msg_match",
    [
        (
            "duplicate_name",
            HOUSE_CATEGORIES,
            [],
            "Stratification name 'duplicate_name' is already used",
        ),
        (
            NAME,
            HOUSE_CATEGORIES + ["slytherin"],
            [],
            f"Found duplicate categories in stratification '{NAME}': ['slytherin']",
        ),
        (
            NAME,
            HOUSE_CATEGORIES + ["gryffindor", "slytherin"],
            [],
            f"Found duplicate categories in stratification '{NAME}': ['gryffindor', 'slytherin']",
        ),
        (
            NAME,
            HOUSE_CATEGORIES,
            ["gryfflepuff"],
            "Excluded categories {'gryfflepuff'} not found in categories",
        ),
    ],
    ids=[
        "duplicate_name",
        "duplicate_category",
        "duplicate_categories",
        "unknown_excluded_category",
    ],
)
def test_add_stratification_raises(
    name: str,
    categories: list[str],
    excluded_categories: list[str],
    msg_match: str,
    mocker: MockerFixture,
) -> None:
    ctx = ResultsContext()
    mocker.patch.object(ctx, "excluded_categories", {name: excluded_categories})
    # Register a stratification to test against duplicate stratifications
    ctx.add_stratification(
        name="duplicate_name",
        requires_columns=["foo"],
        requires_values=[],
        categories=["bar"],
        excluded_categories=None,
        mapper=sorting_hat_serial,
        is_vectorized=False,
    )
    with pytest.raises(ValueError, match=re.escape(msg_match)):
        ctx.add_stratification(
            name=name,
            requires_columns=NAME_COLUMNS,
            requires_values=[],
            categories=categories,
            excluded_categories=excluded_categories,
            mapper=sorting_hat_vectorized,
            is_vectorized=True,
        )


@pytest.mark.parametrize(
    "kwargs",
    [
        {
            "name": "living_person_time",
            "pop_filter": 'alive == "alive" and undead == False',
            "requires_columns": ["alive", "undead"],
            "when": lifecycle_states.COLLECT_METRICS,
        },
        {
            "name": "undead_person_time",
            "pop_filter": "undead == True",
            "requires_columns": ["undead"],
            "when": lifecycle_states.TIME_STEP_PREPARE,
        },
    ],
    ids=["valid_on_collect_metrics", "valid_on_time_step__prepare"],
)
def test_register_observation(kwargs: Any) -> None:
    ctx = ResultsContext()
    assert len(ctx.grouped_observations) == 0
    kwargs["results_formatter"] = lambda: None
    kwargs["stratifications"] = tuple()
    kwargs["aggregator_sources"] = []
    kwargs["aggregator"] = len
    kwargs["requires_values"] = []
    ctx.register_observation(
        observation_type=AddingObservation,
        **kwargs,
    )
    assert len(ctx.grouped_observations) == 1


def test_register_observation_duplicate_name_raises() -> None:
    ctx = ResultsContext()
    ctx.register_observation(
        observation_type=AddingObservation,
        name="some-observation-name",
        pop_filter="some-pop-filter",
        when="some-when",
        requires_columns=[],
        requires_values=[],
        results_formatter=lambda df: df,
        stratifications=(),
        aggregator_sources=[],
        aggregator=len,
    )
    with pytest.raises(
        ValueError, match="Observation name 'some-observation-name' is already used: "
    ):
        # register a different observation but w/ the same name
        ctx.register_observation(
            observation_type=ConcatenatingObservation,
            name="some-observation-name",
            pop_filter="some-other-pop-filter",
            when="some-other-when",
            requires_columns=[],
            requires_values=[],
            stratifications=None,
        )


@pytest.mark.parametrize(
    "aggregator_sources, aggregator, stratifications",
    [
        ([], len, ["house", "familiar"]),
        (["power_level"], sum, ["house", "familiar"]),
        ([], _aggregate_state_person_time, ["house", "familiar"]),
        ([], len, ["house"]),
        (["power_level"], sum, ["house"]),
        ([], _aggregate_state_person_time, ["house"]),
    ],
    ids=[
        "len_aggregator_two_stratifications",
        "sum_aggregator_two_stratifications",
        "custom_aggregator_two_stratifications",
        "len_aggregator_one_stratification",
        "sum_aggregator_one_stratification",
        "custom_aggregator_one_stratification",
    ],
)
def test_adding_observation_gather_results(
    aggregator_sources: list[str],
    aggregator: Callable[..., int | float],
    stratifications: list[str],
    event: Event,
) -> None:
    """Test cases where every stratification is in gather_results. Checks for
    existence and correctness of results"""
    ctx = ResultsContext()

    # Generate population DataFrame
    population = BASE_POPULATION.copy()
    for stratification in stratifications:
        population[get_mapped_col_name(stratification)] = population[stratification]

    # Set up stratifications
    if "house" in stratifications:
        ctx.add_stratification(
            name="house",
            requires_columns=["house"],
            requires_values=[],
            categories=HOUSE_CATEGORIES,
            excluded_categories=None,
            mapper=None,
            is_vectorized=True,
        )
    if "familiar" in stratifications:
        ctx.add_stratification(
            name="familiar",
            requires_columns=["familiar"],
            requires_values=[],
            categories=FAMILIARS,
            excluded_categories=None,
            mapper=None,
            is_vectorized=True,
        )
    pop_filter = "tracked==True"
    observation = ctx.register_observation(
        observation_type=AddingObservation,
        name="foo",
        pop_filter=pop_filter,
        requires_columns=aggregator_sources,
        requires_values=[],
        aggregator_sources=aggregator_sources,
        aggregator=aggregator,
        stratifications=tuple(stratifications),
        when=lifecycle_states.COLLECT_METRICS,
        results_formatter=lambda: None,
    )

    filtered_pop = population.query(pop_filter)
    groups = filtered_pop.groupby(stratifications)
    if aggregator == sum:
        power_level_sums = groups[aggregator_sources].sum().squeeze()
        assert len(power_level_sums.unique()) == 1
        expected_result = power_level_sums.iat[0]
    else:
        group_sizes = groups.size()
        assert len(group_sizes.unique()) == 1
        num_stratifications = group_sizes.iat[0]
        expected_result = (
            num_stratifications if aggregator == len else num_stratifications * 28 / 365.25
        )

    i = 0
    for result, _measure, _updater in ctx.gather_results(
        population, event.name, [observation]
    ):
        assert result is not None
        assert all(
            math.isclose(actual_result, expected_result, rel_tol=0.0001)
            for actual_result in result.values
        )
        i += 1
    assert i == 1


def test_concatenating_observation_gather_results(event: Event) -> None:

    ctx = ResultsContext()

    # Generate population DataFrame
    population = BASE_POPULATION.copy()
    # Mock out some extra columns that would be produced by the manager's _prepare_population() method
    population["current_time"] = pd.Timestamp(year=2045, month=1, day=1, hour=12)
    population["event_step_size"] = timedelta(days=28)
    population["event_time"] = pd.Timestamp(year=2045, month=1, day=1, hour=12) + timedelta(
        days=28
    )

    lifecycle_state = lifecycle_states.COLLECT_METRICS
    pop_filter = "house=='hufflepuff'"
    included_cols = ["familiar", "house"]
    observation = ctx.register_observation(
        observation_type=ConcatenatingObservation,
        name="foo",
        pop_filter=pop_filter,
        when=lifecycle_state,
        requires_columns=included_cols,
        requires_values=[],
        results_formatter=lambda _, __: pd.DataFrame(),
        stratifications=None,
    )

    filtered_pop = population.query(pop_filter)

    i = 0
    for result, _measure, _updater in ctx.gather_results(
        population, event.name, [observation]
    ):
        assert result is not None
        assert result.equals(filtered_pop[["event_time"] + included_cols])
        i += 1
    assert i == 1


@pytest.mark.parametrize(
    "name, aggregator_sources, aggregator, stratifications",
    [
        ("wizard_count", [], len, ["house", "familiar"]),
        ("power_level_total", ["power_level"], sum, ["house", "familiar"]),
        (
            "wizard_time",
            [],
            _aggregate_state_person_time,
            ["house", "familiar"],
        ),
        ("wizard_count", [], len, ["familiar"]),
        ("power_level_total", ["power_level"], sum, ["familiar"]),
        (
            "wizard_time",
            [],
            _aggregate_state_person_time,
            ["familiar"],
        ),
    ],
    ids=[
        "len_aggregator_two_stratifications",
        "sum_aggregator_two_stratifications",
        "custom_aggregator_two_stratifications",
        "len_aggregator_one_stratification",
        "sum_aggregator_one_stratification",
        "custom_aggregator_one_stratification",
    ],
)
def test_gather_results_partial_stratifications_in_results(
    name: str,
    aggregator_sources: list[str],
    aggregator: Callable[..., int | float],
    stratifications: list[str],
    event: Event,
) -> None:
    """Test cases where not all stratifications are observed for gather_results. This looks for existence of
    unobserved stratifications and ensures their values are 0"""
    ctx = ResultsContext()

    # Generate population DataFrame
    population = BASE_POPULATION.copy()
    # Remove an entire category from a stratification
    population = population[population["familiar"] != "unladen_swallow"].reset_index()

    # Set up stratifications
    if "house" in stratifications:
        ctx.add_stratification(
            name="house",
            requires_columns=["house"],
            requires_values=[],
            categories=HOUSE_CATEGORIES,
            excluded_categories=None,
            mapper=None,
            is_vectorized=True,
        )
        population[get_mapped_col_name("house")] = population["house"].astype(
            pd.CategoricalDtype(categories=HOUSE_CATEGORIES, ordered=True)
        )
    if "familiar" in stratifications:
        ctx.add_stratification(
            name="familiar",
            requires_columns=["familiar"],
            requires_values=[],
            categories=FAMILIARS,
            excluded_categories=None,
            mapper=None,
            is_vectorized=True,
        )
        population[get_mapped_col_name("familiar")] = population["familiar"].astype(
            pd.CategoricalDtype(categories=FAMILIARS, ordered=True)
        )

    observation = ctx.register_observation(
        observation_type=AddingObservation,
        name=name,
        pop_filter="tracked==True",
        requires_columns=aggregator_sources,
        requires_values=[],
        aggregator_sources=aggregator_sources,
        aggregator=aggregator,
        stratifications=tuple(stratifications),
        when=lifecycle_states.COLLECT_METRICS,
        results_formatter=lambda: None,
    )

    for results, _measure, _formatter in ctx.gather_results(
        population, event.name, [observation]
    ):
        assert results is not None
        unladen_results = results.reset_index().query('familiar=="unladen_swallow"')
        assert len(unladen_results) > 0
        assert (unladen_results[VALUE_COLUMN] == 0).all()


def test_gather_results_with_empty_pop_filter(event: Event) -> None:
    """Test case where pop_filter filters to an empty population. gather_results
    should return None.
    """
    ctx = ResultsContext()

    # Generate population DataFrame
    population = BASE_POPULATION.copy()

    lifecycle_state = lifecycle_states.COLLECT_METRICS
    observation = ctx.register_observation(
        observation_type=AddingObservation,
        name="wizard_count",
        pop_filter="house == 'durmstrang'",
        requires_columns=["house"],
        requires_values=[],
        aggregator_sources=[],
        aggregator=len,
        stratifications=tuple(),
        when=lifecycle_state,
        results_formatter=lambda: None,
    )

    for result, _measure, _updater in ctx.gather_results(
        population, event.name, [observation]
    ):
        assert not result


def test_gather_results_with_no_stratifications(event: Event) -> None:
    """Test case where we have no stratifications. gather_results should return one value."""
    ctx = ResultsContext()

    # Generate population DataFrame
    population = BASE_POPULATION.copy()

    lifecycle_state = lifecycle_states.COLLECT_METRICS
    observation = ctx.register_observation(
        observation_type=AddingObservation,
        name="wizard_count",
        pop_filter="",
        requires_columns=[],
        requires_values=[],
        aggregator_sources=None,
        aggregator=len,
        stratifications=tuple(),
        when=lifecycle_state,
        results_formatter=lambda: None,
    )

    assert len(ctx.stratifications) == 0
    assert (
        len(
            list(
                result
                for result, _measure, _updater in ctx.gather_results(
                    population, event.name, [observation]
                )
            )
        )
        == 1
    )


def test_bad_aggregator_stratification(event: Event) -> None:
    """Test if an exception gets raised when a stratification that doesn't
    exist is attempted to be used, as expected."""
    ctx = ResultsContext()

    # Generate population DataFrame
    population = BASE_POPULATION.copy()
    lifecycle_state = lifecycle_states.COLLECT_METRICS

    # Set up stratifications
    ctx.add_stratification(
        name="house",
        requires_columns=["house"],
        requires_values=[],
        categories=HOUSE_CATEGORIES,
        excluded_categories=None,
        mapper=None,
        is_vectorized=True,
    )
    ctx.add_stratification(
        name="familiar",
        requires_columns=["familiar"],
        requires_values=[],
        categories=FAMILIARS,
        excluded_categories=None,
        mapper=None,
        is_vectorized=True,
    )
    observation = ctx.register_observation(
        observation_type=AddingObservation,
        name="this_shouldnt_work",
        pop_filter="",
        requires_columns=[],
        requires_values=[],
        aggregator_sources=[],
        aggregator=sum,
        stratifications=("house", "height"),  # `height` is not a stratification
        when=lifecycle_state,
        results_formatter=lambda: None,
    )

    with pytest.raises(KeyError, match="height"):
        for result, _measure, _updater in ctx.gather_results(
            population, event.name, [observation]
        ):
            print(result)


@pytest.mark.parametrize(
    ["lifecycle_state", "time", "expected_observations"],
    [
        (lifecycle_states.COLLECT_METRICS, 0, ["obs1", "obs3"]),
        (lifecycle_states.TIME_STEP_PREPARE, 0, ["obs2"]),
        (lifecycle_states.COLLECT_METRICS, 1, ["obs1"]),
    ],
    ids=["collect_metrics_time_0", "time_step_prepare_time_0", "collect_metrics_time_1"],
)
def test_get_observations(
    lifecycle_state: str, time: int, expected_observations: list[str]
) -> None:
    ctx = ResultsContext()
    register_observation_kwargs = {
        "observation_type": AddingObservation,
        "pop_filter": "",
        "requires_columns": [],
        "requires_values": [],
        "results_formatter": lambda: None,
        "stratifications": (),
        "aggregator_sources": None,
        "aggregator": len,
    }

    ctx.register_observation(
        name="obs1", when=lifecycle_states.COLLECT_METRICS, **register_observation_kwargs  # type: ignore[arg-type]
    )
    ctx.register_observation(
        name="obs2", when=lifecycle_states.TIME_STEP_PREPARE, **register_observation_kwargs  # type: ignore[arg-type]
    )
    ctx.register_observation(
        name="obs3",
        when=lifecycle_states.COLLECT_METRICS,
        to_observe=lambda event: event.time == 0,
        **register_observation_kwargs,  # type: ignore[arg-type]
    )

    event = Event(
        name=lifecycle_state, index=pd.Index([0]), user_data={}, time=time, step_size=1
    )

    assert [obs.name for obs in ctx.get_observations(event)] == expected_observations


@pytest.mark.parametrize("resource_type", ["columns", "values"])
@pytest.mark.parametrize(
    "observation_names, stratification_names, expected_resources",
    [
        (["obs1", "obs2"], ["strat1", "strat2"], {"x", "y", "z", "v"}),
        (["obs3"], ["strat1", "strat2"], {"x", "y", "w", "v"}),
        ([], ["strat1"], {"x", "y"}),
        (["obs2"], [], {"y", "z"}),
        ([], [], set()),
    ],
    ids=[
        "obs_and_strat_with_overlap",
        "obs_and_strat_without_overlap",
        "no_observations",
        "no_stratifications",
        "neither",
    ],
)
def test_get_required_resources(
    resource_type: str,
    observation_names: list[str],
    stratification_names: list[str],
    expected_resources: set[str],
) -> None:
    ctx = ResultsContext()

    all_observations = {}
    register_observation_kwargs = {
        "observation_type": AddingObservation,
        "pop_filter": "",
        "when": lifecycle_states.COLLECT_METRICS,
        "results_formatter": lambda: None,
        "stratifications": (),
        "aggregator_sources": None,
        "aggregator": len,
    }

    def get_required_resources_kwargs(
        resource_type: str, resources: list[str]
    ) -> dict[str, list[str] | list[Pipeline]]:
        if resource_type == "columns":
            return {"requires_columns": resources, "requires_values": []}
        elif resource_type == "values":
            return {
                "requires_values": [Pipeline(r) for r in resources],
                "requires_columns": [],
            }
        else:
            raise ValueError(f"Unknown resource_type: {resource_type}")

    all_observations["obs1"] = ctx.register_observation(
        name="obs1",
        **get_required_resources_kwargs(resource_type, ["x", "y"]),  # type: ignore[arg-type]
        **register_observation_kwargs,  # type: ignore[arg-type]
    )
    all_observations["obs2"] = ctx.register_observation(
        name="obs2",
        **get_required_resources_kwargs(resource_type, ["y", "z"]),  # type: ignore[arg-type]
        **register_observation_kwargs,  # type: ignore[arg-type]
    )
    all_observations["obs3"] = ctx.register_observation(
        name="obs3",
        **get_required_resources_kwargs(resource_type, ["w"]),  # type: ignore[arg-type]
        **register_observation_kwargs,  # type: ignore[arg-type]
    )

    all_stratifications = {}
    stratification_kwargs = {
        "categories": ["cat1", "cat2"],
        "excluded_categories": ["cat3"],
        "mapper": lambda df: df["a"] + df["b"],
        "is_vectorized": True,
    }
    all_stratifications["strat1"] = Stratification(
        name="strat1",
        **get_required_resources_kwargs(resource_type, ["x", "y"]),  # type: ignore[arg-type]
        **stratification_kwargs,  # type: ignore[arg-type]
    )
    all_stratifications["strat2"] = Stratification(
        name="strat2",
        **get_required_resources_kwargs(resource_type, ["x", "v"]),  # type: ignore[arg-type]
        **stratification_kwargs,  # type: ignore[arg-type]
    )

    observations = [all_observations[name] for name in observation_names]
    stratifications = [all_stratifications[name] for name in stratification_names]

    if resource_type == "columns":
        actual_columns = ctx.get_required_columns(observations, stratifications)
        assert set(actual_columns) == {"tracked"} | expected_resources
    elif resource_type == "values":
        actual_columns = [
            p.name for p in ctx.get_required_values(observations, stratifications)
        ]
        assert set(actual_columns) == expected_resources
    else:
        raise ValueError(f"Unknown resource_type: {resource_type}")


@pytest.mark.parametrize(
    "pop_filter, stratifications",
    [
        ('familiar=="cat"', tuple()),
        ('familiar=="spaghetti_yeti"', tuple()),
        ("", ("new_col1",)),
        ("", ("new_col1", "new_col2")),
        ('familiar=="cat"', ("new_col1",)),
        ("", tuple()),
    ],
    ids=[
        "pop_filter",
        "pop_filter_empties_dataframe",
        "single_excluded_stratification",
        "two_excluded_stratifications",
        "pop_filter_and_excluded_stratification",
        "no_pop_filter_or_excluded_stratifications",
    ],
)
def test__filter_population(pop_filter: str, stratifications: tuple[str, ...]) -> None:
    population = BASE_POPULATION.copy()
    if stratifications:
        # Make some of the stratifications missing to mimic mapping to excluded categories
        population["new_col1"] = "new_value1"
        population.loc[population["tracked"] == True, "new_col1"] = np.nan
        if len(stratifications) == 2:
            population["new_col2"] = "new_value2"
            population.loc[population["new_col1"].notna(), "new_col2"] = np.nan
        # Add on the post-stratified columns
        for stratification in stratifications:
            mapped_col = f"{stratification}_mapped_values"
            population[mapped_col] = population[stratification]

    filtered_pop = ResultsContext()._filter_population(
        population=population, pop_filter=pop_filter, stratification_names=stratifications
    )
    expected = population.copy()
    if pop_filter:
        familiar = pop_filter.split("==")[1].strip('"')
        expected = expected[expected["familiar"] == familiar]
    for stratification in stratifications:
        expected = expected[expected[stratification].notna()]
    assert filtered_pop.equals(expected)


@pytest.mark.parametrize(
    "stratifications, values",
    [
        (("familiar",), [FAMILIARS]),
        (("familiar", "house"), [FAMILIARS, HOUSE_CATEGORIES]),
        ((), "foo"),
    ],
)
def test__get_groups(stratifications: tuple[str, ...], values: str | list[list[str]]) -> None:
    filtered_pop = BASE_POPULATION.copy()
    # Generate the post-stratified columns
    for stratification in stratifications:
        mapped_col = f"{stratification}_mapped_values"
        filtered_pop[mapped_col] = filtered_pop[stratification]
    groups = ResultsContext()._get_groups(
        stratifications=stratifications, filtered_pop=filtered_pop
    )
    assert isinstance(groups, DataFrameGroupBy)
    if stratifications:
        combinations = set(itertools.product(*values))
        if len(values) == 1:
            # convert from set of tuples to set of strings
            combinations = set(comb[0] for comb in combinations)
        # Check that all familiars exist
        assert set(groups.groups.keys()) == combinations
        # Check that the entire population is included
        assert sum([len(value) for value in groups.groups.values()]) == len(BASE_POPULATION)
    else:
        item = groups.groups.popitem()
        # Check that there are no other groups
        assert not groups.groups
        # Check that the group is 'all' and includes the entire population
        key, val = item
        assert key == "all"
        assert val.equals(BASE_POPULATION.index)



from typing import Any

import pytest
from pytest_mock import MockerFixture

from tests.helpers import MockComponentA, MockComponentB, MockGenericComponent, MockManager
from vivarium import Component
from vivarium.exceptions import VivariumError
from vivarium.framework.components.manager import (
    ComponentConfigError,
    ComponentManager,
    OrderedComponentSet,
)
from vivarium.framework.configuration import build_simulation_configuration
from vivarium.manager import Manager


def test_component_set_add() -> None:
    component_list = OrderedComponentSet()

    component_0 = MockComponentA(name="component_0")
    component_list.add(component_0)

    component_1 = MockComponentA(name="component_1")
    component_list.add(component_1)

    # duplicates by name
    with pytest.raises(ComponentConfigError, match="duplicate name"):
        component_list.add(component_0)


def test_component_set_update() -> None:
    component_list = OrderedComponentSet()

    components = [MockComponentA(name="component_0"), MockComponentA("component_1")]

    component_list.update(components)

    with pytest.raises(ComponentConfigError, match="duplicate name"):
        component_list.update(components)


def test_component_set_initialization() -> None:
    component_1 = MockComponentA()
    component_2 = MockComponentB()

    component_list = OrderedComponentSet(component_1, component_2)
    assert component_list.components == [component_1, component_2]


def test_component_set_pop() -> None:
    component = MockComponentA()
    component_list = OrderedComponentSet(component)

    c = component_list.pop()
    assert c == component

    with pytest.raises(IndexError):
        component_list.pop()


def test_component_set_contains() -> None:
    component_list = OrderedComponentSet()

    assert not bool(component_list)
    assert len(component_list) == 0

    component_1 = MockComponentA()
    component_2 = MockComponentB()
    component_3 = MockComponentA(name="absent")
    component_list = OrderedComponentSet(component_1, component_2)

    assert component_1 in component_list
    assert component_3 not in component_list
    with pytest.raises(ComponentConfigError, match="no name"):
        _ = 10 in component_list  # type: ignore[operator]


def test_component_set_eq() -> None:
    component_1 = MockComponentA()
    component_2 = MockComponentB()
    component_list = OrderedComponentSet(component_1, component_2)

    assert component_list == component_list
    assert component_list != 10

    second_list = OrderedComponentSet(component_1)
    assert component_list != second_list


def test_component_set_bool_len() -> None:
    component_list = OrderedComponentSet()

    assert not bool(component_list)
    assert len(component_list) == 0

    component_1 = MockComponentA()
    component_2 = MockComponentB()
    component_list = OrderedComponentSet(component_1, component_2)

    assert bool(component_list)
    assert len(component_list) == 2


def test_component_set_dunder_add() -> None:
    l1 = OrderedComponentSet(*[MockComponentA(name=str(i)) for i in range(5)])
    l2 = OrderedComponentSet(*[MockComponentA(name=str(i)) for i in range(5, 10)])
    combined = OrderedComponentSet(*[MockComponentA(name=str(i)) for i in range(10)])

    assert l1 + l2 == combined


def test_manager_init() -> None:
    m = ComponentManager()
    assert not m._managers
    assert not m._components
    with pytest.raises(VivariumError, match="ComponentManager has no configuration tree"):
        _ = m.configuration
    assert m.name == "component_manager"
    assert repr(m) == str(m) == "ComponentManager()"


def test_manager_get_file() -> None:
    mock_component = MockGenericComponent("foo")
    # Extract the full path to where MockGenericComponent is defined
    mock_component_path = (
        __file__.split("/tests/")[0]
        + "/"
        + MockGenericComponent.__module__.replace(".", "/")
        + ".py"
    )
    assert ComponentManager._get_file(mock_component) == mock_component_path
    mock_component.__module__ = "__main__"
    assert ComponentManager._get_file(mock_component) == "__main__"


def test_flatten_simple() -> None:
    components = [MockComponentA(name=str(i)) for i in range(10)]
    assert ComponentManager._flatten(components) == components


def test_flatten_with_lists() -> None:
    components = []
    for i in range(5):
        for j in range(5):
            components.append(MockComponentA(name=str(5 * i + j)))
    out = ComponentManager._flatten(components)
    expected = [MockComponentA(name=str(i)) for i in range(25)]
    assert out == expected


def test_flatten_with_sub_components() -> None:
    components = []
    for i in range(5):
        name, *args = [str(5 * i + j) for j in range(5)]
        components.append(MockComponentB(*args, name=name))
    out = ComponentManager._flatten(components)
    expected = [MockComponentB(name=str(i)) for i in range(25)]
    assert out == expected


def test_flatten_with_nested_sub_components() -> None:
    def nest(start: int, depth: int) -> Component:
        if depth == 1:
            return MockComponentA(name=str(start))
        c = MockComponentA(name=str(start))
        c._sub_components = [nest(start + 1, depth - 1)]
        return c

    components: list[Component] = []
    for i in range(5):
        components.append(nest(5 * i, 5))
    out = ComponentManager._flatten(components)
    expected = [MockComponentA(name=str(i)) for i in range(25)]
    assert out == expected

    # Lists with nested subcomponents
    out = ComponentManager._flatten([components, components])
    assert out == 2 * expected


def test_setup_components(mocker: MockerFixture) -> None:
    builder = mocker.Mock()
    builder.configuration = {}
    mocker.patch("vivarium.framework.results.observer.Observer.set_results_dir")
    mocker.patch("vivarium.framework.results.observer.Observer.get_configuration")
    mock_a = MockComponentA("test_a")
    mock_b = MockComponentB("test_b")
    components = OrderedComponentSet(mock_a, mock_b)
    ComponentManager._setup_components(builder, components)

    assert mock_a.builder_used_for_setup is None  # class has no setup method
    assert mock_b.builder_used_for_setup is builder


def test_apply_configuration_defaults() -> None:
    config = build_simulation_configuration()
    c = config.to_dict()
    cm = ComponentManager()
    cm._configuration = config
    component = MockGenericComponent("test_component")

    cm.apply_configuration_defaults(component)
    c.update(component.configuration_defaults)
    assert config.to_dict() == c


def test_apply_configuration_defaults_no_op() -> None:
    config = build_simulation_configuration()
    cm = ComponentManager()
    cm._configuration = config
    component = MockComponentA()

    cm.apply_configuration_defaults(component)
    assert config.to_dict() == component.configuration_defaults


def test_apply_configuration_defaults_duplicate() -> None:
    config = build_simulation_configuration()

    cm = ComponentManager()
    cm._configuration = config
    component = MockGenericComponent("test_component")

    cm.apply_configuration_defaults(component)
    cm._components.add(component)
    with pytest.raises(ComponentConfigError, match="but it has already been set"):
        cm.apply_configuration_defaults(component)


def test_apply_configuration_defaults_bad_structure() -> None:
    class BadConfigComponent(MockComponentA):
        @property
        def configuration_defaults(self) -> dict[str, Any]:
            return {"test_component": "val"}

    config = build_simulation_configuration()

    cm = ComponentManager()
    cm._configuration = config
    component1 = MockGenericComponent("test_component")
    component2 = BadConfigComponent(name="test_component2")

    cm.apply_configuration_defaults(component1)
    cm._components.add(component1)
    with pytest.raises(ComponentConfigError, match="attempting to alter the structure"):
        cm.apply_configuration_defaults(component2)


def test_add_components() -> None:
    config = build_simulation_configuration()
    cm = ComponentManager()
    cm._configuration = config
    components: list[Component] = [MockGenericComponent(f"component_{i}") for i in range(5)]
    cm.add_components(components)
    assert cm._components == OrderedComponentSet(*components)
    for component in components:
        assert (
            config[component.name].to_dict()
            == component.configuration_defaults[component.name]
        )
    assert cm.list_components() == {c.name: c for c in components}


def test_add_managers() -> None:
    cm = ComponentManager()
    cm._configuration = build_simulation_configuration()
    assert not cm._managers
    mock_managers: list[Manager] = [MockManager(f"manager_{i}") for i in range(5)]
    cm.add_managers(mock_managers)
    assert cm._managers == OrderedComponentSet(*mock_managers)


@pytest.mark.parametrize(
    "components",
    ([MockComponentA("Eric"), MockComponentB("half", "a", "bee")], [MockComponentA("Eric")]),
)
def test_component_manager_add_components(components: list[Component]) -> None:
    config = build_simulation_configuration()
    cm = ComponentManager()
    cm._configuration = config
    mock_managers: list[Manager] = [
        MockManager(f"{component.name}_manager") for component in components
    ]
    cm.add_managers(mock_managers)
    assert cm._managers == OrderedComponentSet(*mock_managers)

    config = build_simulation_configuration()
    cm = ComponentManager()
    cm._configuration = config
    cm.add_components(components)
    assert cm._components == OrderedComponentSet(*ComponentManager._flatten(components))


@pytest.mark.parametrize(
    "components",
    (
        [MockComponentA(), MockComponentA()],
        [MockComponentA(), MockComponentA(), MockComponentB("foo", "bar")],
    ),
)
def test_component_manager_add_components_duplicated(components: list[Component]) -> None:
    config = build_simulation_configuration()
    cm = ComponentManager()
    cm._configuration = config
    mock_managers: list[Manager] = [
        MockManager(f"{component.name}_manager") for component in components
    ]
    with pytest.raises(
        ComponentConfigError,
        match=f"Attempting to add a component with duplicate name: {MockComponentA()}",
    ):
        cm.add_managers(mock_managers)

    config = build_simulation_configuration()
    cm = ComponentManager()
    cm._configuration = config
    with pytest.raises(
        ComponentConfigError,
        match="is attempting to set the configuration value mock_component_a, but it has already been set by mock_component_a",
    ):
        cm.add_components(components)



"""
==========================
Results System Exceptions
==========================

Errors related to the results system
"""

from vivarium.exceptions import VivariumError


class ResultsConfigurationError(VivariumError):
    """Error raised when the results stratifications are improperly configured."""

    pass



import pandas as pd
import pytest

from vivarium.framework.event import Event
from vivarium.framework.lifecycle.entities import LifeCycle, LifeCyclePhase, LifeCycleState
from vivarium.framework.lifecycle.exceptions import ConstraintError, LifeCycleError
from vivarium.framework.lifecycle.lifecycle_states import INITIALIZATION
from vivarium.framework.lifecycle.manager import LifeCycleManager


def test_state_add_next() -> None:
    a = LifeCycleState("a")
    b = LifeCycleState("b")
    c = LifeCycleState("c")

    assert a.valid_next_state(None)
    a.add_next(b)
    assert a.valid_next_state(b)
    assert not a.valid_next_state(c)
    assert not a.valid_next_state(None)

    assert b.valid_next_state(None)
    b.add_next(a, loop=True)
    assert b.valid_next_state(a)
    assert b.valid_next_state(None)
    b.add_next(c)
    assert not b.valid_next_state(None)
    assert b.valid_next_state(a)
    assert b.valid_next_state(c)

    assert c.valid_next_state(None)
    assert not c.valid_next_state(a)
    assert not c.valid_next_state(b)


def test_state_entrance_count() -> None:
    a = LifeCycleState("a")
    assert a.entrance_count == 0
    a.enter()
    a.enter()
    assert a.entrance_count == 2


class BumbleBee:
    def __init__(self, name: str) -> None:
        self.name = name

    def buzz(self, _event: Event) -> None:
        pass

    def what_we_doin(self, _event: Event) -> None:
        pass

    def __call__(self) -> str:
        return "On the way!"


def sting(_event: Event) -> None:
    pass


def test_state_add_handlers() -> None:
    a = LifeCycleState("a")
    b = BumbleBee("Alice")
    a.add_handlers([b.buzz, b.what_we_doin, sting])
    assert a._handlers == [
        "BumbleBee(Alice).buzz",
        "BumbleBee(Alice).what_we_doin",
        "Unbound function sting",
    ]


def test_phase_init_no_loop() -> None:
    phase_name = "test"
    state_names = ["a", "b", "c", "d", "e"]

    p = LifeCyclePhase(phase_name, state_names, loop=False)
    assert p.name == phase_name
    assert [s.name for s in p.states] == state_names
    current = p.states[0]
    for state in p.states[1:]:
        assert current.valid_next_state(state)
        current = state
    assert current.valid_next_state(None)


def test_phase_init_loop() -> None:
    phase_name = "test"
    state_names = ["a", "b", "c", "d", "e"]

    p = LifeCyclePhase(phase_name, state_names, loop=True)
    assert p.name == phase_name
    assert [s.name for s in p.states] == state_names
    current = p.states[0]
    for state in p.states[1:]:
        assert current.valid_next_state(state)
        current = state
    assert current.valid_next_state(p.states[0])


def test_phase_add_next() -> None:
    phase1_name = "test1"
    phase2_name = "test2"
    states1 = ["a", "b", "c", "d", "e"]
    states2 = ["1", "2", "3", "4", "5"]

    p1 = LifeCyclePhase(phase1_name, states1, loop=True)
    p2 = LifeCyclePhase(phase2_name, states2, loop=False)

    p1.add_next(p2)
    assert p1.states[-1].valid_next_state(p1.states[0])
    assert p1.states[-1].valid_next_state(p2.states[0])


def test_phase_get_state() -> None:
    name = "test"
    state_names = ["a", "b", "c", "d", "e"]

    p = LifeCyclePhase(name, state_names, loop=False)

    for name, state in zip(state_names, p.states):
        assert state is p.get_state(name)

    with pytest.raises(IndexError):
        p.get_state("not_a_state")


def test_phase_contains() -> None:
    name = "test"
    state_names = ["a", "b", "c", "d", "e"]

    p = LifeCyclePhase(name, state_names, loop=False)

    for name in state_names:
        assert name in p

    for name in ["1", "2", "3", ""]:
        assert name not in p


def test_lifecycle_init() -> None:
    lc = LifeCycle()
    assert "initialization" in lc
    assert "not_a_state" not in lc


def test_lifecycle_validate() -> None:
    lc = LifeCycle()

    with pytest.raises(LifeCycleError, match="phase names must be unique"):
        lc._validate("initialization", ["a", "b", "c"])

    with pytest.raises(LifeCycleError, match="duplicate state names"):
        lc._validate("new_phase", ["a", "b", "c", "b"])

    with pytest.raises(LifeCycleError, match=INITIALIZATION):
        lc._validate("new_phase", ["a", "b", INITIALIZATION])

    lc._validate("new_phase", ["a", "b", "c"])


@pytest.mark.parametrize("loop", [True, False])
def test_lifecycle_add_phase_fail(loop: bool) -> None:
    lc = LifeCycle()

    with pytest.raises(LifeCycleError, match="phase names must be unique"):
        lc.add_phase("initialization", ["a", "b", "c"], loop)

    with pytest.raises(LifeCycleError, match="duplicate state names"):
        lc.add_phase("new_phase", ["a", "b", "c", "b"], loop)

    with pytest.raises(LifeCycleError, match=INITIALIZATION):
        lc.add_phase("new_phase", ["a", "b", INITIALIZATION], loop)


@pytest.mark.parametrize("loop", [True, False])
def test_lifecycle_add_phase(loop: bool) -> None:
    lc = LifeCycle()
    lc.add_phase("phase1", ["a", "b", "c"], loop)


def test_lifecycle_get_state() -> None:
    lc = LifeCycle()
    assert lc.get_state(INITIALIZATION).name == INITIALIZATION

    with pytest.raises(LifeCycleError, match="non-existent state"):
        lc.get_state("not_a_state")


def test_lifecycle_get_states() -> None:
    lc = LifeCycle()
    assert lc.get_state_names(INITIALIZATION) == [INITIALIZATION]

    with pytest.raises(LifeCycleError, match="non-existent phase"):
        lc.get_state_names("not_a_phase")


def test_lifecycle_integration() -> None:
    lc = LifeCycle()

    with pytest.raises(LifeCycleError, match="non-existent state"):
        lc.get_state("a")
    with pytest.raises(LifeCycleError, match="non-existent phase"):
        lc.get_state_names("phase1")

    init = lc.get_state(INITIALIZATION)
    assert init.valid_next_state(None)

    lc.add_phase("phase1", ["a"], loop=False)
    a = lc.get_state("a")

    assert lc.get_state_names("phase1") == ["a"]
    assert init.valid_next_state(a)
    assert not init.valid_next_state(None)
    assert a.valid_next_state(None)

    lc.add_phase("phase2", ["b"], loop=True)
    b = lc.get_state("b")

    assert lc.get_state_names("phase2") == ["b"]
    assert a.valid_next_state(b)
    assert b.valid_next_state(None)
    assert b.valid_next_state(b)

    lc.add_phase("phase3", ["c", "d"], loop=True)
    c = lc.get_state("c")
    d = lc.get_state("d")

    assert not b.valid_next_state(None)
    assert b.valid_next_state(b)
    assert b.valid_next_state(c)
    assert c.valid_next_state(d)
    assert d.valid_next_state(c)
    assert d.valid_next_state(None)

    with pytest.raises(LifeCycleError, match="phase names must be unique"):
        lc.add_phase("phase2", ["1", "2"], loop=True)

    with pytest.raises(LifeCycleError, match="duplicate state names"):
        lc.add_phase("phase4", ["1", "2", "3", "2"], loop=False)

    with pytest.raises(LifeCycleError, match="state names must be unique"):
        lc.add_phase("phase4", ["a", "f", "g"], loop=False)

    with pytest.raises(LifeCycleError, match="non-existent state"):
        lc.get_state("not a state")

    with pytest.raises(LifeCycleError, match="non-existent phase"):
        lc.get_state_names("not a phase")


def test_manager_init() -> None:
    lm = LifeCycleManager()
    assert lm.name == "life_cycle_manager"
    assert lm.current_state == INITIALIZATION


@pytest.mark.parametrize("loop", [True, False])
def test_lifecycle_manager_add_phase_fail(loop: bool) -> None:
    lm = LifeCycleManager()

    with pytest.raises(LifeCycleError, match="phase names must be unique"):
        lm.add_phase("initialization", ["a", "b", "c"], loop)

    with pytest.raises(LifeCycleError, match="duplicate state names"):
        lm.add_phase("new_phase", ["a", "b", "c", "b"], loop)

    with pytest.raises(LifeCycleError, match=INITIALIZATION):
        lm.add_phase("new_phase", ["a", "b", INITIALIZATION], loop)


@pytest.mark.parametrize("loop", [True, False])
def test_lifecycle_manager_add_phase(loop: bool) -> None:
    lm = LifeCycleManager()
    lm.add_phase("phase1", ["a", "b", "c"], loop)


def test_lifecycle_manager_get_states() -> None:
    lm = LifeCycleManager()
    assert lm.get_state_names("initialization") == [INITIALIZATION]

    with pytest.raises(LifeCycleError, match="non-existent phase"):
        lm.get_state_names("not_a_phase")


def test_lifecycle_manager_set_state() -> None:
    lm = LifeCycleManager()
    lm.add_phase("phase1", ["a", "b"])
    lm.add_phase("phase2", ["c", "d"])

    assert lm.current_state == INITIALIZATION
    lm.set_state("a")
    assert lm.current_state == "a"
    lm.set_state("b")
    assert lm.current_state == "b"

    with pytest.raises(LifeCycleError, match="Invalid transition"):
        lm.set_state("a")  # phase1 does not loop

    with pytest.raises(LifeCycleError, match="Invalid transition"):
        lm.set_state("d")  # next state is c

    with pytest.raises(LifeCycleError, match="non-existent state"):
        lm.set_state("e")

    lm.set_state("c")
    assert lm.current_state == "c"
    lm.set_state("d")
    assert lm.current_state == "d"


def test_lifecycle_manager_set_state_with_loop() -> None:
    lm = LifeCycleManager()
    lm.add_phase("phase1", ["a", "b"], loop=True)
    lm.add_phase("phase2", ["c", "d"])

    assert lm.current_state == INITIALIZATION
    lm.set_state("a")
    assert lm.current_state == "a"
    lm.set_state("b")
    assert lm.current_state == "b"

    lm.set_state("a")  # phase permits loops
    assert lm.current_state == "a"
    lm.set_state("b")
    assert lm.current_state == "b"

    with pytest.raises(LifeCycleError, match="Invalid transition"):
        lm.set_state("d")  # next state is a or c

    with pytest.raises(LifeCycleError, match="non-existent state"):
        lm.set_state("e")

    lm.set_state("c")
    assert lm.current_state == "c"
    lm.set_state("d")
    assert lm.current_state == "d"

    with pytest.raises(LifeCycleError, match="Invalid transition"):
        lm.set_state("c")  # phase 2 does not permit loops


def test_lifecycle_manager_add_handlers() -> None:
    lm = LifeCycleManager()
    lm.add_phase("phase1", ["a"])

    init = lm.lifecycle.get_state(INITIALIZATION)
    a = lm.lifecycle.get_state("a")

    b = BumbleBee("Alice")

    lm.add_handlers(INITIALIZATION, [b.buzz, b.what_we_doin, sting])
    assert init._handlers == [
        "BumbleBee(Alice).buzz",
        "BumbleBee(Alice).what_we_doin",
        "Unbound function sting",
    ]

    lm.add_handlers("a", [b.buzz])
    assert a._handlers == ["BumbleBee(Alice).buzz"]


def test_lifecycle_manager_add_constraint_fail() -> None:
    lm = LifeCycleManager()
    lm.add_phase("phase1", ["a", "b", "c"])

    alice = BumbleBee("Alice")

    with pytest.raises(ValueError, match="Must provide exactly one"):
        lm.add_constraint(alice.buzz)

    with pytest.raises(ValueError, match="Must provide exactly one"):
        lm.add_constraint(
            alice.buzz, allow_during=[INITIALIZATION, "a"], restrict_during=["c"]
        )

    with pytest.raises(LifeCycleError, match="states not in the life cycle"):
        lm.add_constraint(alice.buzz, allow_during=["not_a_state"])

    with pytest.raises(LifeCycleError, match="states not in the life cycle"):
        lm.add_constraint(alice.buzz, restrict_during=["not_a_state"])

    lm.add_constraint(alice.buzz, allow_during=["a"])

    with pytest.raises(ConstraintError, match="already been constrained"):
        lm.add_constraint(alice.buzz, allow_during=["b"])

    with pytest.raises(TypeError, match="bound object methods"):
        lm.add_constraint(test_lifecycle_manager_add_constraint_fail, allow_during=["a"])

    with pytest.raises(ValueError, match="normal object methods"):
        lm.add_constraint(alice.__call__, allow_during=["a"])


def test_lifecycle_manager_add_constraint() -> None:
    lm = LifeCycleManager()
    lm.add_phase("phase1", ["a", "b", "c"], loop=True)

    alice = BumbleBee("Alice")
    bob = BumbleBee("Bob")

    lm.add_constraint(alice.buzz, allow_during=[INITIALIZATION, "a"])
    lm.add_constraint(bob.buzz, restrict_during=[INITIALIZATION, "c"])

    useless_event = Event(
        name="useless_event",
        index=pd.Index([0]),
        user_data={},
        time=0,
        step_size=1,
    )

    alice.buzz(useless_event)
    with pytest.raises(ConstraintError, match="it may only be called during"):
        bob.buzz(useless_event)

    lm.set_state("a")
    alice.buzz(useless_event)
    bob.buzz(useless_event)

    lm.set_state("b")
    with pytest.raises(ConstraintError, match="it may only be called during"):
        alice.buzz(useless_event)
    bob.buzz(useless_event)

    lm.set_state("c")
    with pytest.raises(ConstraintError, match="it may only be called during"):
        alice.buzz(useless_event)
    with pytest.raises(ConstraintError, match="it may only be called during"):
        bob.buzz(useless_event)

    lm.set_state("a")
    alice.buzz(useless_event)
    bob.buzz(useless_event)



"""
================
Lifecycle States
================

This module defines constants representing the states in the simulation lifecycle.

"""


# Initialization states
INITIALIZATION = "initialization"
"""The name of the initialization state in the simulation lifecycle."""

# Setup states
SETUP = "setup"
"""The name of the setup state in the simulation lifecycle."""

POST_SETUP = "post_setup"
"""The name of the post-setup state in the simulation lifecycle."""

POPULATION_CREATION = "population_creation"
"""The name of the population creation state in the simulation lifecycle."""

#  Main loop states
TIME_STEP_PREPARE = "time_step__prepare"
"""The name of the time step preparation state in the simulation lifecycle."""
TIME_STEP = "time_step"
"""The name of the time step state in the simulation lifecycle."""
TIME_STEP_CLEANUP = "time_step__cleanup"
"""The name of the time step cleanup state in the simulation lifecycle."""
COLLECT_METRICS = "collect_metrics"
"""The name of the collect metrics state in the simulation lifecycle."""

# Simulation end states
SIMULATION_END = "simulation_end"
"""The name of the simulation end state in the simulation lifecycle."""
REPORT = "report"
"""The name of the report state in the simulation lifecycle."""



"""
=====================
Life Cycle Management
=====================

The life cycle is a representation of the flow of execution states in a
:mod:`vivarium` simulation. The tools in this model allow a simulation to
formally represent its execution state and use the formal representation to
enforce run-time contracts.

There are two flavors of contracts that this system enforces:

 - **Constraints**: These are contracts around when certain methods,
   particularly those available off the :ref:`Builder <builder_concept>`,
   can be used. For example, :term:`simulants <Simulant>` should only be
   added to the simulation during initial population creation and during
   the main simulation loop, otherwise services necessary for initializing
   that population's attributes may not exist. By applying a constraint,
   we can provide very clear errors about what went wrong, rather than
   a deep and unintelligible stack trace.
 - **Ordering Contracts**: The
   :class:`~vivarium.framework.engine.SimulationContext` will construct
   the formal representation of the life cycle during its initialization.
   Once generated, the context declares as it transitions between
   different lifecycle states and the tools here ensure that only valid
   transitions occur.  These kinds of contracts are particularly useful
   during interactive usage, as they prevent users from, for example,
   running a simulation whose population has not been created.

The tools here also allow for introspection of the simulation life cycle.

"""

from vivarium.framework.lifecycle.exceptions import ConstraintError, LifeCycleError
from vivarium.framework.lifecycle.interface import LifeCycleInterface
from vivarium.framework.lifecycle.manager import LifeCycleManager



"""
==========================
Vivarium Interactive Tools
==========================

This module provides an interface for interactive simulation usage. The main
part is the :class:`InteractiveContext`, a sub-class of the main simulation
object in ``vivarium`` that has been extended to include convenience
methods for running and exploring the simulation in an interactive setting.

See the associated tutorials for :ref:`running <interactive_tutorial>` and
:ref:`exploring <exploration_tutorial>` for more information.

"""
from __future__ import annotations

from collections.abc import Callable
from math import ceil
from typing import Any

import pandas as pd

from vivarium.framework.engine import SimulationContext
from vivarium.framework.event import Event
from vivarium.framework.values import Pipeline
from vivarium.interface.utilities import log_progress, run_from_ipython
from vivarium.types import ClockStepSize, ClockTime


class InteractiveContext(SimulationContext):
    """A simulation context with helper methods for running simulations interactively."""

    def __init__(self, *args: Any, setup: bool = True, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

        if setup:
            self.setup()

    @property
    def current_time(self) -> ClockTime:
        """Returns the current simulation time."""
        return self._clock.time

    def setup(self) -> None:
        super().setup()
        self.initialize_simulants()

    def step(self, step_size: ClockStepSize | None = None) -> None:
        """Advance the simulation one step.

        Parameters
        ----------
        step_size
            An optional size of step to take. Must be compatible with the
            simulation clock's step size (usually a pandas.Timedelta).
        """
        old_step_size = self._clock._clock_step_size
        if step_size is not None:
            if not (
                isinstance(step_size, type(self._clock.step_size))
                or isinstance(self._clock.step_size, type(step_size))
            ):
                raise ValueError(
                    f"Provided time must be compatible with {type(self._clock.step_size)}"
                )
            self._clock._clock_step_size = step_size
        super().step()
        self._clock._clock_step_size = old_step_size

    def run(self, with_logging: bool = True) -> None:  # type: ignore [override]
        """Run the simulation for the duration specified in the configuration.

        Parameters
        ----------
        with_logging
            Whether or not to log the simulation steps. Only works in an ipython
            environment.

        Returns
        -------
            The number of steps the simulation took.
        """
        self.run_until(self._clock.stop_time, with_logging=with_logging)

    def run_for(self, duration: ClockStepSize, with_logging: bool = True) -> None:
        """Run the simulation for the given time duration.

        Parameters
        ----------
        duration
            The length of time to run the simulation for. Should be compatible
            with the simulation clock's step size (usually a pandas
            Timedelta).
        with_logging
            Whether or not to log the simulation steps. Only works in an ipython
            environment.

        Returns
        -------
            The number of steps the simulation took.
        """
        self.run_until(self._clock.time + duration, with_logging=with_logging)  # type: ignore [operator]

    def run_until(self, end_time: ClockTime, with_logging: bool = True) -> None:
        """Run the simulation until the provided end time.

        Parameters
        ----------
        end_time
            The time to run the simulation until. The simulation will run until
            its clock is greater than or equal to the provided end time. Must be
            compatible with the simulation clock's step size (usually a pandas.Timestamp)

        with_logging
            Whether or not to log the simulation steps. Only works in an ipython
            environment.

        Returns
        -------
            The number of steps the simulation took.
        """
        if not (
            isinstance(end_time, type(self._clock.time))
            or isinstance(self._clock.time, type(end_time))
        ):
            raise ValueError(
                f"Provided time must be compatible with {type(self._clock.time)}"
            )

        iterations = int(ceil((end_time - self._clock.time) / self._clock.step_size))  # type: ignore [operator, arg-type]
        self.take_steps(number_of_steps=iterations, with_logging=with_logging)
        assert self._clock.time - self._clock.step_size < end_time <= self._clock.time  # type: ignore [operator]
        print("Simulation complete after", iterations, "iterations")

    def take_steps(
        self,
        number_of_steps: int = 1,
        step_size: ClockStepSize | None = None,
        with_logging: bool = True,
    ) -> None:
        """Run the simulation for the given number of steps.

        Parameters
        ----------
        number_of_steps
            The number of steps to take.
        step_size
            An optional size of step to take. Must be compatible with the
            simulation clock's step size (usually a pandas.Timedelta).
        with_logging
            Whether or not to log the simulation steps. Only works in an ipython
            environment.
        """
        if not isinstance(number_of_steps, int):
            raise ValueError("Number of steps must be an integer.")

        if run_from_ipython() and with_logging:
            for _ in log_progress(range(number_of_steps), name="Step"):
                self.step(step_size)
        else:
            for _ in range(number_of_steps):
                self.step(step_size)

    def get_population(self, untracked: bool = False) -> pd.DataFrame:
        """Get a copy of the population state table.

        Parameters
        ----------
        untracked
            Whether or not to return simulants who are no longer being tracked
            by the simulation.

        Returns
        -------
            The population state table.
        """
        return self._population.get_population(untracked)

    def list_values(self) -> list[str]:
        """List the names of all pipelines in the simulation."""
        return list(self._values.keys())

    def get_value(self, value_pipeline_name: str) -> Pipeline:
        """Get the value pipeline associated with the given name."""
        if value_pipeline_name not in self.list_values():
            raise ValueError(f"No value pipeline '{value_pipeline_name}' registered.")
        return self._values.get_value(value_pipeline_name)

    def list_events(self) -> list[str]:
        """List all event types registered with the simulation."""
        return self._events.list_events()

    def get_listeners(self, event_name: str) -> dict[int, list[Callable[[Event], None]]]:
        """Get all listeners of a particular type of event.

        Available events can be found by calling
        :func:`InteractiveContext.list_events`.

        Parameters
        ----------
        event_name
            The name of the event to grab the listeners for.

        Returns
        -------
            A dictionary that maps each priority level of the named event's
            listeners to a list of listeners at that level.
        """
        if event_name not in self._events:
            raise ValueError(f"No event {event_name} in system.")
        return self._events.get_listeners(event_name)

    def get_emitter(
        self, event_name: str
    ) -> Callable[[pd.Index[int], dict[str, Any] | None], Event]:
        """Get the callable that emits the given type of events.

        Available events can be found by calling
        :func:`InteractiveContext.list_events`.

        Parameters
        ----------
        event_name
            The name of the event to grab the listeners for.

        Returns
        -------
            The callable that emits the named event.
        """
        if event_name not in self._events:
            raise ValueError(f"No event {event_name} in system.")
        return self._events.get_emitter(event_name)

    def list_components(self) -> dict[str, Any]:
        """Get a mapping of component names to components currently in the simulation.

        Returns
        -------
            A dictionary mapping component names to components.
        """
        return self._component_manager.list_components()

    def get_component(self, name: str) -> Any:
        """Get the component in the simulation that has ``name``, if present.
        Names are guaranteed to be unique.

        Parameters
        ----------
        name
            A component name.

        Returns
        -------
            A component that has the name ``name`` else None.
        """
        return self._component_manager.get_component(name)

    def print_initializer_order(self) -> None:
        """Print the order in which population initializers are called."""
        initializers = []
        for r in self._resource.get_population_initializers():
            name = r.__name__
            if hasattr(r, "__self__"):
                obj = r.__self__
                initializers.append(f"{obj.__class__.__name__}({obj.name}).{name}")
            else:
                initializers.append(f"Unbound function {name}")
        print("\n".join(initializers))

    def print_lifecycle_order(self) -> None:
        """Print the order of lifecycle events (including user event handlers)."""
        print(self._lifecycle)

    def __repr__(self) -> str:
        return "InteractiveContext()"



"""
===========================
Vivarium Command Line Tools
===========================

``vivarium`` provides the tool :command:`simulate` for running simulations
from the command line.  It provides three subcommands:

.. list-table:: ``simulate`` sub-commands
    :header-rows: 1
    :widths: 30, 40

    *   - Name
        - Description
    *   - | **run**
        - | Runs a single simulation from a model specification file.
    *   - | **test**
        - | Runs an example simulation that comes packaged with ``vivarium``.
          | Useful as an installation test.
    *   - | **profile**
        - | Produces a profile of a simulation using the python
          | :mod:`cProfile` module

For more information, see the :ref:`tutorial <cli_tutorial>` on running
simulations from the command line.

.. click:: vivarium.interface.cli:simulate
   :prog: simulate
   :show-nested:

"""

import cProfile
import os
import pstats
from datetime import datetime as dt
from pathlib import Path
from time import time

import click
import yaml
from loguru import logger

from vivarium.examples import disease_model
from vivarium.framework.engine import SimulationContext
from vivarium.framework.logging import (
    configure_logging_to_file,
    configure_logging_to_terminal,
)
from vivarium.framework.utilities import handle_exceptions
from vivarium.interface.utilities import get_output_root


@click.group()
def simulate() -> None:
    """A command line utility for running a single simulation.

    You may initiate a new run with the ``run`` sub-command, initiate a test
    run of a provided model specification with the ``test`` subcommand, or
    profile a simulation run with the ``profile`` subcommand.
    """
    pass


@simulate.command()
@click.argument(
    "model_specification", type=click.Path(exists=True, dir_okay=False, resolve_path=True)
)
@click.option(
    "--artifact_path",
    "-i",
    type=click.Path(resolve_path=True),
    help="The path to the artifact data file.",
)
@click.option(
    "--results_directory",
    "-o",
    type=click.Path(resolve_path=True),
    default=Path("~/vivarium_results/").expanduser(),
    help="The directory to write results to. A folder will be created "
    "in this directory with the same name as the configuration file.",
)
@click.option(
    "--verbose",
    "-v",
    is_flag=True,
    help="Logs verbosely. Useful for debugging and development.",
)
@click.option(
    "--quiet",
    "-q",
    is_flag=True,
    help="Suppresses all logging except for warnings and errors.",
)
@click.option(
    "--pdb",
    "with_debugger",
    is_flag=True,
    help="Drop into python debugger if an error occurs.",
)
def run(
    model_specification: Path,
    artifact_path: Path,
    results_directory: Path,
    verbose: bool,
    quiet: bool,
    with_debugger: bool,
) -> None:
    """Run a simulation from the command line.

    The simulation itself is defined by the given MODEL_SPECIFICATION yaml file.

    Within the results directory, which defaults to ~/vivarium_results if none
    is provided, a subdirectory will be created with the same name as the
    MODEL_SPECIFICATION if one does not exist. Results will be written to a
    further subdirectory named after the start time of the simulation run.
    """
    if verbose and quiet:
        raise click.UsageError("Cannot be both verbose and quiet.")
    verbosity = 1 + int(verbose) - int(quiet)
    configure_logging_to_terminal(verbosity=verbosity, long_format=False)

    start = time()

    results_root = get_output_root(results_directory, model_specification, artifact_path)
    # Update permissions mask (assign to variable to avoid printing previous value)
    _ = os.umask(0o002)
    results_root.mkdir(parents=True, exist_ok=False)
    output_data_root = results_root / "results"
    output_data_root.mkdir(parents=True, exist_ok=False)

    configure_logging_to_file(output_directory=results_root)

    output_data = {"results_directory": str(output_data_root)}
    input_data = {}
    if artifact_path:
        input_data["artifact_path"] = artifact_path
    override_configuration = {"output_data": output_data, "input_data": input_data}

    sim = SimulationContext(
        model_specification=model_specification, configuration=override_configuration
    )
    with open(f"{results_directory}/model_specification.yaml", "w") as f:
        yaml.dump(sim.model_specification.to_dict(), f)

    main = handle_exceptions(sim.run_simulation, logger, with_debugger)
    main()

    # Save out simulation metadata
    metadata = {}
    metadata["random_seed"] = sim.configuration.randomness.random_seed
    metadata["input_draw"] = sim.configuration.input_data.input_draw_number
    metadata["simulation_run_time"] = time() - start
    metadata["artifact_path"] = artifact_path
    with open(results_root / "metadata.yaml", "w") as f:
        yaml.dump(metadata, f, default_flow_style=False)

    logger.info(f"Simulation finished.\nResults written to {str(output_data_root)}")


@simulate.command()
def test() -> None:
    """Run a test simulation using the ``disease_model.yaml`` model specification
    provided in the examples directory.
    """
    configure_logging_to_terminal(verbosity=2, long_format=False)
    model_specification = disease_model.get_model_specification_path()

    sim = SimulationContext(model_specification)

    main = handle_exceptions(sim.run_simulation, logger, with_debugger=False)
    main()

    click.echo()
    click.secho("Installation test successful!", fg="green")


@simulate.command()
@click.argument(
    "model_specification",
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
)
@click.option(
    "--results_directory",
    "-o",
    type=click.Path(resolve_path=True),
    default=Path("~/vivarium_results/").expanduser(),
    show_default=True,
    help=(
        "The directory to write results to. A folder will be created "
        "in this directory with the same name as the configuration file."
    ),
)
@click.option(
    "--skip_writing",
    is_flag=True,
    help=(
        "Skip writing the simulation results to the output directory; the time spent "
        "normally writing simulation results to disk will not be included in the profiling "
        "statistics."
    ),
)
@click.option(
    "--skip_processing",
    is_flag=True,
    help=(
        "Skip processing the resulting binary file to a human-readable .txt file "
        "sorted by cumulative runtime; the resulting .stats file can still be read "
        "and processed later using the pstats module."
    ),
)
def profile(
    model_specification: Path,
    results_directory: Path,
    skip_writing: bool,
    skip_processing: bool,
) -> None:
    """Run a simulation based on the provided MODEL_SPECIFICATION and profile the run."""
    model_specification = Path(model_specification)
    results_directory = Path(results_directory)
    results_root = results_directory / f"{dt.now().strftime('%Y_%m_%d_%H_%M_%S')}"
    configure_logging_to_file(output_directory=results_root)

    if skip_writing:
        configuration_override = {}
    else:
        output_data_root = results_root / "results"
        output_data_root.mkdir(parents=True, exist_ok=False)
        configuration_override = {
            "output_data": {"results_directory": str(output_data_root)},
        }

    out_stats_file = results_root / f"{model_specification.name}".replace("yaml", "stats")
    sim = SimulationContext(model_specification, configuration=configuration_override)
    command = f"sim.run_simulation()"
    cProfile.runctx(command, globals=globals(), locals=locals(), filename=str(out_stats_file))

    if not skip_processing:
        out_txt_file = Path(str(out_stats_file) + ".txt")
        with out_txt_file.open("w") as f:
            p = pstats.Stats(str(out_stats_file), stream=f)
            p.sort_stats("cumulative")
            p.print_stats()



"""
======================
Lifecycle Exceptions
======================

Exception classes for the lifecycle management system.

"""
from __future__ import annotations

from vivarium.exceptions import VivariumError


class LifeCycleError(VivariumError):
    """Generic error class for the life cycle management system."""

    pass


class InvalidTransitionError(LifeCycleError):
    """Error raised when life cycle ordering contracts are violated."""

    pass


class ConstraintError(LifeCycleError):
    """Error raised when life cycle constraint contracts are violated."""

    pass



from typing import Any

import pytest
from layered_config_tree.main import LayeredConfigTree
from pytest_mock import MockerFixture

from vivarium import InteractiveContext
from vivarium.framework.components.manager import ComponentConfigError
from vivarium.framework.engine import Builder
from vivarium.framework.results.observer import Observer


class TestObserver(Observer):
    def register_observations(self, builder: Builder) -> None:
        pass


class TestDefaultObserverStratifications(Observer):
    def register_observations(self, builder: Builder) -> None:
        pass


class TestObserverStratifications(Observer):
    @property
    def configuration_defaults(self) -> dict[str, Any]:
        return {
            "stratification": {
                self.get_configuration_name(): {
                    "exclude": ["baz"],
                    "include": ["foo"],
                },
            },
        }

    def register_observations(self, builder: Builder) -> None:
        pass


def test_observer_instantiation() -> None:
    observer = TestObserver()
    assert observer.name == "test_observer"


@pytest.mark.parametrize(
    "is_interactive, results_dir",
    [
        (False, "/some/results/dir"),
        (True, None),
    ],
)
def test_set_results_dir(
    is_interactive: bool, results_dir: str | None, mocker: MockerFixture
) -> None:
    builder = mocker.Mock()
    if is_interactive:
        builder.configuration = LayeredConfigTree()
    else:
        builder.configuration = LayeredConfigTree(
            {
                "output_data": {"results_directory": results_dir},
            }
        )

    observer = TestObserver()
    observer.set_results_dir(builder)

    assert observer.results_dir == results_dir


def test_observer_get_configuration(
    base_config: LayeredConfigTree,
) -> None:

    observer = TestObserverStratifications()
    sim = InteractiveContext(
        base_config,
        components=[observer],
    )
    sim_observer_config = sim.configuration["stratification"][
        observer.get_configuration_name()
    ]
    # Observer.configuration calls get_configuration
    observer_config = observer.configuration
    assert observer_config is not None
    assert observer_config.to_dict() == dict(sim_observer_config)


def test_duplicated_observer_error(base_config: LayeredConfigTree) -> None:
    observer1 = TestObserverStratifications()
    observer2 = TestObserverStratifications()
    with pytest.raises(
        ComponentConfigError,
        match="is attempting to set the configuration value test, but it has already been set",
    ):
        InteractiveContext(
            base_config,
            components=[observer1, observer2],
        )



from pathlib import Path

import pytest

from vivarium.interface.utilities import get_output_model_name_string

_MODEL_SPEC_STEM = "model_spec_name"
_ARTIFACT_STEM = "artifact_name"
_ARTIFACT_FROM_MODEL_SPEC_STEM = f"{_ARTIFACT_STEM}_from_model_spec"

_ARTIFACT_PATH = Path(f"/totally/fake/path/for/artifact/{_ARTIFACT_STEM}.hdf")
_MODEL_SPEC_ARTIFACT_PATH = Path(
    f"/totally/fake/path/for/artifact/{_ARTIFACT_FROM_MODEL_SPEC_STEM}.hdf"
)

_MODEL_SPEC_CONTENTS_WITHOUT = """
configuration:
    input_data:
        input_draw_number: 0
"""

_MODEL_SPEC_CONTENTS_WITH = (
    _MODEL_SPEC_CONTENTS_WITHOUT
    + f"""
        artifact_path: '{_MODEL_SPEC_ARTIFACT_PATH}'
"""
)


def _write_file(path: Path, contents: str) -> None:
    with open(path, "w") as file:
        file.write(contents)


@pytest.mark.parametrize(
    "artifact_path, model_spec_filename, contents, expected_output",
    [
        (
            # Given an input artifact path, use that
            _ARTIFACT_PATH,
            f"{_MODEL_SPEC_STEM}_with.yaml",
            _MODEL_SPEC_CONTENTS_WITH,
            _ARTIFACT_STEM,
        ),
        (
            # Without an input artifact path, but given model spec with artifact, use that
            None,
            f"{_MODEL_SPEC_STEM}_with.yaml",
            _MODEL_SPEC_CONTENTS_WITH,
            _ARTIFACT_FROM_MODEL_SPEC_STEM,
        ),
        (
            # Without the things in previous parameters, choose the stem of the model spec
            None,
            f"{_MODEL_SPEC_STEM}.yaml",
            _MODEL_SPEC_CONTENTS_WITHOUT,
            _MODEL_SPEC_STEM,
        ),
    ],
)
def test_get_output_model_name_string(
    artifact_path: Path,
    model_spec_filename: str,
    contents: str,
    expected_output: str,
    tmp_path: Path,
) -> None:
    model_spec_path = Path(f"{tmp_path}/{model_spec_filename}")
    _write_file(model_spec_path, contents)

    output = get_output_model_name_string(artifact_path, model_spec_path)

    assert output == expected_output



import json
import sys

from packaging.version import parse

with open("python_versions.json", "r") as f:
    supported_python_versions = json.load(f)

python_versions = [parse(v) for v in supported_python_versions]
min_version = min(python_versions)
max_version = max(python_versions)

if not (
    min_version <= parse(".".join([str(v) for v in sys.version_info[:2]])) <= max_version
):
    # Python 3.5 does not support f-strings
    py_version = ".".join([str(v) for v in sys.version_info[:3]])
    error = (
        "\n----------------------------------------\n"
        "Error: Vivarium runs under python {min_version}-{max_version}.\n"
        "You are running python {py_version}".format(
            min_version=min_version, max_version=max_version, py_version=py_version
        )
    )
    print(error, file=sys.stderr)
    sys.exit(1)

from pathlib import Path

from setuptools import find_packages, setup

if __name__ == "__main__":
    base_dir = Path(__file__).parent
    src_dir = base_dir / "src"

    about = {}
    with (src_dir / "vivarium" / "__about__.py").open() as f:
        exec(f.read(), about)

    with (base_dir / "README.rst").open() as f:
        long_description = f.read()

    install_requirements = [
        "vivarium_dependencies[numpy_lt_2,pandas,pyyaml,scipy,click,tables,loguru,pyarrow,networkx]",
        "vivarium_build_utils>=2.0.1,<3.0.0",
        "layered_config_tree",
        "dill",
    ]

    setup_requires = ["setuptools_scm"]

    interactive_requirements = [
        "vivarium_dependencies[interactive]",
    ]

    test_requirements = [
        "vivarium_dependencies[pytest]",
        "vivarium_testing_utils",
    ]

    lint_requirements = [
        "vivarium_dependencies[lint]",
        "mypy<1.17.0",  # FIXME [MIC-6218]
    ]

    doc_requirements = [
        "vivarium_dependencies[sphinx,sphinx-click,ipython,matplotlib]",
        "sphinxcontrib-video",
    ]

    setup(
        name=about["__title__"],
        description=about["__summary__"],
        long_description=long_description,
        license=about["__license__"],
        url=about["__uri__"],
        author=about["__author__"],
        author_email=about["__email__"],
        classifiers=[
            "Intended Audience :: Developers",
            "Intended Audience :: Education",
            "Intended Audience :: Science/Research",
            "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
            "Natural Language :: English",
            "Operating System :: MacOS :: MacOS X",
            "Operating System :: POSIX",
            "Operating System :: POSIX :: BSD",
            "Operating System :: POSIX :: Linux",
            "Operating System :: Microsoft :: Windows",
            "Programming Language :: Python",
            "Programming Language :: Python :: Implementation :: CPython",
            "Topic :: Education",
            "Topic :: Scientific/Engineering",
            "Topic :: Scientific/Engineering :: Artificial Life",
            "Topic :: Scientific/Engineering :: Mathematics",
            "Topic :: Scientific/Engineering :: Medical Science Apps.",
            "Topic :: Scientific/Engineering :: Physics",
            "Topic :: Software Development :: Libraries",
        ],
        package_dir={"": "src"},
        packages=find_packages(where="src"),
        include_package_data=True,
        install_requires=install_requirements,
        tests_require=test_requirements,
        extras_require={
            "docs": doc_requirements,
            "test": test_requirements,
            "interactive": interactive_requirements,
            "dev": doc_requirements
            + test_requirements
            + lint_requirements
            + interactive_requirements,
        },
        entry_points="""
                [console_scripts]
                simulate=vivarium.interface.cli:simulate
            """,
        zip_safe=False,
        use_scm_version={
            "write_to": "src/vivarium/_version.py",
            "write_to_template": '__version__ = "{version}"\n',
            "tag_regex": r"^(?P<prefix>v)?(?P<version>[^\+]+)(?P<suffix>.*)?$",
        },
        setup_requires=setup_requires,
    )



"""
================
Constraint Maker
================

Factory for making state-based constraints on component methods.

"""
from __future__ import annotations

import functools
from collections.abc import Callable
from typing import TYPE_CHECKING, Any

from vivarium.framework.lifecycle.exceptions import ConstraintError

if TYPE_CHECKING:
    from vivarium.framework.lifecycle.manager import LifeCycleManager


class ConstraintMaker:
    """Factory for making state-based constraints on component methods."""

    def __init__(self, lifecycle_manager: LifeCycleManager):
        self.lifecycle_manager = lifecycle_manager
        self.constraints: set[str] = set()

    def check_valid_state(
        self, method: Callable[..., Any], permitted_states: list[str]
    ) -> None:
        """Ensures a component method is being called during an allowed state.

        Parameters
        ----------
        method
            The method the constraint is applied to.
        permitted_states
            The states in which the method is permitted to be called.

        Raises
        ------
        ConstraintError
            If the method is being called outside the permitted states.
        """
        current_state = self.lifecycle_manager.current_state
        if current_state not in permitted_states:
            raise ConstraintError(
                f"Trying to call {method} during {current_state},"
                f" but it may only be called during {permitted_states}."
            )

    def constrain_normal_method(
        self, method: Callable[..., Any], permitted_states: list[str]
    ) -> Callable[..., Any]:
        """Only permit a method to be called during the provided states.

        Constraints are applied by dynamically wrapping and binding a method
        to an existing component at run time.

        Parameters
        ----------
        method
            The method to constrain.
        permitted_states
            The life cycle states in which the method can be called.

        Returns
        -------
            The constrained method.
        """

        @functools.wraps(method)
        def _wrapped(*args: Any, **kwargs: Any) -> Any:
            self.check_valid_state(method, permitted_states)
            # Call the __func__ because we're rebinding _wrapped to the method
            # name on the object.  If we called method directly, we'd get
            # two copies of self.
            return method.__func__(*args, **kwargs)  # type: ignore [attr-defined]

        # Invoke the descriptor protocol to bind the wrapped method to the
        # component instance.
        rebound_method: Callable[..., Any] = _wrapped.__get__(method.__self__, method.__self__.__class__)  # type: ignore [attr-defined]
        # Then update the instance dictionary to reflect that the wrapped
        # method is bound to the original name.
        setattr(method.__self__, method.__name__, rebound_method)  # type: ignore [attr-defined]
        return rebound_method

    @staticmethod
    def to_guid(method: Callable[..., Any]) -> str:
        """Convert a method on to a global id.

        Because we dynamically rebind methods, the old ones will get garbage
        collected, making :func:`id` unreliable for checking if a method
        has been constrained before.

        Parameters
        ----------
        method
            The method to convert to a global id.

        Returns
        -------
            The global id of the method.
        """
        return f"{method.__self__.name}.{method.__name__}"  # type: ignore [attr-defined]

    def __call__(
        self, method: Callable[..., Any], permitted_states: list[str]
    ) -> Callable[..., Any]:
        """Only permit a method to be called during the provided states.

        Constraints are applied by dynamically wrapping and binding a method
        to an existing component at run time.

        Parameters
        ----------
        method
            The method to constrain.
        permitted_states
            The life cycle states in which the method can be called.

        Returns
        -------
            The constrained method.

        Raises
        ------
        TypeError
            If an unbound function is supplied for constraint.
        ValueError
            If the provided method is a python "special" method (i.e. a
            method surrounded by double underscores).
        """
        if not hasattr(method, "__self__"):
            raise TypeError(
                "Can only apply constraints to bound object methods. "
                f"You supplied the function {method}."
            )
        name = method.__name__
        if name.startswith("__") and name.endswith("__"):
            raise ValueError(
                "Can only apply constraints to normal object methods. "
                f" You supplied {method}."
            )

        if self.to_guid(method) in self.constraints:
            raise ConstraintError(f"Method {method} has already been constrained.")

        self.constraints.add(self.to_guid(method))
        return self.constrain_normal_method(method, permitted_states)



"""
==================
Lifecycle Entities
==================

Core entity classes for the lifecycle management system.

"""
from __future__ import annotations

import textwrap
from collections.abc import Callable
from typing import TYPE_CHECKING

from vivarium.framework.lifecycle.exceptions import LifeCycleError
from vivarium.framework.lifecycle.lifecycle_states import INITIALIZATION

if TYPE_CHECKING:
    from vivarium.framework.event import Event


class LifeCycleState:
    """A representation of a simulation run state."""

    def __init__(self, name: str):
        self._name = name
        self._next: LifeCycleState | None = None
        self._loop_next: LifeCycleState | None = None
        self._entrance_count = 0
        self._handlers: list[str] = []

    @property
    def name(self) -> str:
        """The name of the lifecycle state."""
        return self._name

    @property
    def entrance_count(self) -> int:
        """The number of times this state has been entered."""
        return self._entrance_count

    def add_next(self, next_state: LifeCycleState, loop: bool = False) -> None:
        """Link this state to the next state in the simulation life cycle.

        States are linked together and used to ensure that the simulation
        life cycle proceeds in the proper order.  A life cycle state can be
        bound to two ``next`` states to allow for loops in the life cycle and
        both are considered valid when checking for valid state transitions.
        The first represents the linear progression through the simulation,
        while the second represents a loop in the life cycle.

        Parameters
        ----------
        next_state
            The next state in the simulation life cycle.
        loop
            Whether the provided state is the linear next state or a loop
            back to a previous state in the life cycle.
        """
        if loop:
            self._loop_next = next_state
        else:
            self._next = next_state

    def valid_next_state(self, state: LifeCycleState | None) -> bool:
        """Check if the provided state is valid for a life cycle transition.

        Parameters
        ----------
        state
            The state to check.

        Returns
        -------
            Whether the state is valid for a transition.
        """
        return (state is None and state is self._next) or (
            state is not None and (state is self._next or state is self._loop_next)
        )

    def enter(self) -> None:
        """Marks an entrance into this state."""
        self._entrance_count += 1

    def add_handlers(self, handlers: list[Callable[[Event], None]]) -> None:
        """Registers a set of functions that will be executed during the state.

        The primary use case here is for introspection and reporting.
        For setting constraints, see
        :meth:`vivarium.framework.lifecycle.interface.LifeCycleInterface.add_constraint`.

        Parameters
        ----------
        handlers
            The set of functions that will be executed during this state.
        """
        for h in handlers:
            name = h.__name__
            if hasattr(h, "__self__"):
                obj = h.__self__
                self._handlers.append(f"{obj.__class__.__name__}({obj.name}).{name}")
            else:
                self._handlers.append(f"Unbound function {name}")

    def __repr__(self) -> str:
        return f"LifeCycleState(name={self.name})"

    def __str__(self) -> str:
        return "\n\t".join([self.name] + self._handlers)


class LifeCyclePhase:
    """A representation of a distinct lifecycle phase in the simulation.

    A lifecycle phase is composed of one or more unique lifecycle states.
    There is exactly one state within the phase which serves as a valid
    exit point from the phase.  The states may operate in a loop.

    """

    def __init__(self, name: str, states: list[str], loop: bool):
        self._name = name
        self._states = [LifeCycleState(states[0])]
        self._loop = loop
        for s in states[1:]:
            self._states.append(LifeCycleState(s))
            self._states[-2].add_next(self._states[-1])
        if self._loop:
            self._states[-1].add_next(self._states[0], loop=True)

    @property
    def name(self) -> str:
        """The name of this life cycle phase."""
        return self._name

    @property
    def states(self) -> tuple[LifeCycleState, ...]:
        """The states in this life cycle phase in order of execution."""
        return tuple(self._states)

    def add_next(self, phase: LifeCyclePhase) -> None:
        """Link the provided phase as the next phase in the life cycle."""
        self._states[-1].add_next(phase._states[0])

    def get_state(self, state_name: str) -> LifeCycleState:
        """Retrieve a life cycle state by name from the phase."""
        return [s for s in self._states if s.name == state_name].pop()

    def __contains__(self, state_name: str) -> bool:
        return bool([s for s in self._states if s.name == state_name])

    def __repr__(self) -> str:
        return f"LifeCyclePhase(name={self.name}, states={[s.name for s in self.states]})"

    def __str__(self) -> str:
        out = self.name
        if self._loop:
            out += "*"
        out += "\n" + textwrap.indent("\n".join([str(state) for state in self.states]), "\t")
        return out


class LifeCycle:
    """A concrete representation of the flow of simulation execution states."""

    def __init__(self) -> None:
        self._state_names: set[str] = set()
        self._phase_names: set[str] = set()
        self._phases: list[LifeCyclePhase] = []
        self.add_phase("initialization", [INITIALIZATION], loop=False)

    def add_phase(self, phase_name: str, states: list[str], loop: bool) -> None:
        """Add a new phase to the lifecycle.

        Phases must be added in order.

        Parameters
        ----------
        phase_name
            The name of the phase to add.  Phase names must be unique.
        states
            The list of names (in order) of the states that make up the
            life cycle phase.  State names must be unique across the entire
            life cycle.
        loop
            Whether the life cycle phase states loop.

        Raises
        ------
        LifeCycleError
            If the phase or state names are non-unique.
        """
        self._validate(phase_name, states)

        new_phase = LifeCyclePhase(phase_name, states, loop)
        if self._phases:
            self._phases[-1].add_next(new_phase)

        self._state_names.update(states)
        self._phase_names.add(phase_name)
        self._phases.append(new_phase)

    def get_state(self, state_name: str) -> LifeCycleState:
        """Retrieve a life cycle state from the life cycle.

        Parameters
        ----------
        state_name
            The name of the state to retrieve

        Returns
        -------
            The requested state.

        Raises
        ------
        LifeCycleError
            If the requested state does not exist.
        """
        if state_name not in self:
            raise LifeCycleError(f"Attempting to look up non-existent state {state_name}.")
        phase = [p for p in self._phases if state_name in p].pop()
        return phase.get_state(state_name)

    def get_state_names(self, phase_name: str) -> list[str]:
        """Retrieve the names of all states in the provided phase.

        Parameters
        ----------
        phase_name
            The name of the phase to retrieve the state names from.

        Return
        ------
            The state names in the provided phase.

        Raises
        ------
        LifeCycleError
            If the phase does not exist in the life cycle.
        """
        if phase_name not in self._phase_names:
            raise LifeCycleError(
                f"Attempting to look up states from non-existent phase {phase_name}."
            )
        phase = [p for p in self._phases if p.name == phase_name].pop()
        return [s.name for s in phase.states]

    def _validate(self, phase_name: str, states: list[str]) -> None:
        """Validates that a phase and set of states are unique."""
        if phase_name in self._phase_names:
            raise LifeCycleError(
                f"Lifecycle phase names must be unique. You're attempting "
                f"to add {phase_name} but it already exists."
            )
        if len(states) != len(set(states)):
            raise LifeCycleError(
                f"Attempting to create a life cycle phase with duplicate state names. "
                f"States: {states}"
            )
        duplicates = self._state_names.intersection(states)
        if duplicates:
            raise LifeCycleError(
                f"Lifecycle state names must be unique.  You're attempting "
                f"to add {duplicates} but they already exist."
            )

    def __contains__(self, state_name: str) -> bool:
        return state_name in self._state_names

    def __repr__(self) -> str:
        return f"LifeCycle(phases={self._phase_names})"

    def __str__(self) -> str:
        return "\n".join([str(phase) for phase in self._phases])
